library nui_auth;
import 'package:flutter/cupertino.dart';
import 'package:google_sign_in/google_sign_in.dart';
import 'package:nui_auth/domain/nui_user_activity.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/model/interfaces.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_web/nui_web.dart';
import 'domain/nui_auth_token.dart';
import 'domain/nui_user_activity.dart';
import 'web/nui_auth_web_json.dart';
import 'web/nui_auth_api.dart';
import 'result/nui_google_login_result.dart';
import 'result/nui_auth_result.dart';

typedef NUITokenExpiredListener = void Function();

abstract class NUIAuthInterface{
  Future<DataResult<NUIAuthSignUpResult>> signup({@required String username, @required String password, String email});
  Future<DataResult> verifyEmail({@required String code});
  Future<DataResult> resendEmailVerificationCode();
  Future<DataResult<NUIAuthLoginResult>> login({@required String username, @required String password});
  Future<DataResult<NUIGoogleLoginResult>> loginWithGoogle(BuildContext context);
  Future<DataResult> logoutFromGoogle(BuildContext context);
  Future<DataResult> changePassword({@required String oldPassword, @required String newPassword, @required String confirmPassword});
  Future<DataResult> forgotPassword({@required String email});
  NUIAuthAccessToken getAccessToken();
  String getRefreshToken();
  String getUserId();
  Future<DataResult> logout();
  bool isLoggedIn();
  Future<DataResult> logUserActivity({@required NUIUserActivity userActivity});
}

class NUIAuth implements NUIAuthInterface{
  static const String GRANT_TYPE_PASSWORD = "password";
  static const String GRANT_TYPE_REFRESH_TOKEN = "refresh_token";
  static const String _PREF_ACCESS_TOKEN = "istudio_access_token";
  static const String _PREF_REFRESH_TOKEN = "istudio_refresh_token";
  static const String _PREF_EXPIRES_IN = "istudio_expires_in";
  static const String _PREF_SSO_ACCESS_TOKEN = "istudio_sso_access_token";
  static const String _PREF_SIGNUP_TOKEN = "istudio_signup_token";
  static const String _PREF_APP_INSTANCE_ID = "istudio_app_instance_id";
  static const String _MODULE = "NUIAuth";

  String _baseUrl;
  NUIAuthApi _api;
  static NUIAuth _instance;
  GoogleSignIn _googleSignIn;
  NUIAuth._builder(NUIAuthBuilder builder){
    _baseUrl = builder._baseUrl;
    _api = NUIAuthApi();
    _googleSignIn = GoogleSignIn(
      scopes: [
        'email',
        'https://www.googleapis.com/auth/contacts.readonly',
      ],
    );
    _instance = this;
  }

  static NUIAuth get(){
    if(_instance == null) throwNUI(_MODULE, "Instance has not been created with builder");
    return _instance;
  }

  void _saveLoginResponseData(NUIAuthLoginResp body) async{
    final save1 = NUISharedPref.get().save(_PREF_ACCESS_TOKEN, body.accessToken);
    final save2 = NUISharedPref.get().save(_PREF_REFRESH_TOKEN, body.refreshToken);
    final save3 = NUISharedPref.get().save(_PREF_EXPIRES_IN, body.expiresIn);
    await save1;
    await save2;
    await save3;
  }

  @override
  Future<DataResult> changePassword({String oldPassword, String newPassword, String confirmPassword}) async{
    final req = NUIAuthChangePasswordReq(oldPassword: oldPassword, newPassword: newPassword, confirmPassword: confirmPassword);
    final result = await _api.changePassword(request: req);
    return DataResult(success: result.success, message: result.message);
  }

  @override
  Future<DataResult> forgotPassword({String email}) async{
    final req = NUIAuthForgotPasswordReq(email: email);
    final result = await _api.forgotPassword(request: req);
    return DataResult(success: result.success, message: result.message);
  }

  @override
  NUIAuthAccessToken getAccessToken(){
    final token = NUISharedPref.get().quickGetString(_PREF_ACCESS_TOKEN);
    return NUIAuthAccessToken(token: token, refreshToken: getRefreshToken());
  }

  @override
  String getRefreshToken() {
    return NUISharedPref.get().quickGetString(_PREF_REFRESH_TOKEN);
  }

  @override
  String getUserId() {
    return getAccessToken()?.getUserId();
  }

  @override
  bool isLoggedIn() {
    return getAccessToken()?.isExpired(leeway: 0) == false;
  }

  @override
  Future<DataResult> logUserActivity({NUIUserActivity userActivity}) async{
    final result = await _api.logUserActivity(activity: userActivity);
    return DataResult(success: result.success, message: result.message);
  }

  @override
  Future<DataResult<NUIAuthLoginResult>> login({@required String username, @required String password}) async{
    final req = NUIAuthLoginReq(username: username, password: password, grantType: GRANT_TYPE_PASSWORD);
    final result = await _api.login(request: req);
    if(result.success && result.data != null) {
      final authResp = result.data;
      return DataResult(success: result.success, message: result.message, data: NUIAuthLoginResult(accessToken: authResp.accessToken, refreshToken: authResp.refreshToken, expiresIn: authResp.expiresIn));
    }
    else{
      return DataResult(success: result.success, message: result.message);
    }
  }

  @override
  Future<DataResult<NUIGoogleLoginResult>> loginWithGoogle(BuildContext context) async{
    try {
      final loginResult = await _googleSignIn.signIn();
      final bool isSignedIn =  await _googleSignIn.isSignedIn();
      if(isSignedIn){
        final curUser = _googleSignIn.currentUser;
        final authentication = await curUser.authentication;
        final token = NUIAuthAccessToken(token: authentication.accessToken);
        final isExpired = token.isExpired();
        final googleResp = NUIGoogleLoginResult(displayName: curUser.displayName, id: curUser.id, email: curUser.email, photoUrl: curUser.photoUrl, idToken: authentication.idToken, serverAuthCode: authentication.serverAuthCode, isExpired: isExpired);

        // send the token back to iStudio server
        final ssoReq = NUIAuthSSOSignUpReq(provider: NUIAuthSSOSignUpReq.SSO_TYPE_GOOGLE, providerToken: authentication.idToken, email: curUser.email);
        final result = await _api.ssoLogin(request: ssoReq);

        if(result.success){
          // save the 3rd-party access token
          await NUISharedPref.get().save(_PREF_SSO_ACCESS_TOKEN, authentication.idToken);
          // save iStudio access token
          _saveLoginResponseData(result.data);
        }

        return DataResult<NUIGoogleLoginResult>(success: true, message: "Login successful", data: googleResp);
      }
      else{
        return DataResult<NUIGoogleLoginResult>(success: false, message: "Login with Google was unsuccessful");
      }
    } catch (error) {
      logNUI(_MODULE, "Failed to login with Google with error : $error");
      return DataResult(success: false, message: "Failed to login with Google");
    }
  }

  @override
  Future<DataResult> logout() async{
    // delete access token
    await NUISharedPref.get().remove(_PREF_ACCESS_TOKEN);
    final result = await _api.logout();
    return DataResult(success: result.success, message: result.message);
  }

  @override
  Future<DataResult> logoutFromGoogle(BuildContext context) async{
    // sign out from google
    try {
      final result = await _googleSignIn.signOut();
      return DataResult(success: true, message: "Logout from Google successful");
    }catch(e){
      logNUI(_MODULE, "Logout from Google failed with error : $e");
      return DataResult(success: false, message: "Failed to logout from Google");
    }
  }

  @override
  Future<DataResult> resendEmailVerificationCode() async {
    final token = await NUISharedPref.get().getString(_PREF_SIGNUP_TOKEN);

    if (token.isNullOrEmpty()) {
      throwNUI(_MODULE, "Invalid sign up token detected");
      return DataResult(success: false, message: "Invalid sign up token detected");
    }

    final req = NUIAuthVerifyEmailReq(token: token);
    final result = await _api.resendEmailVerificationCode(request: req);
    return DataResult(success: result.success, message: result.message);
  }

  @override
  Future<DataResult<NUIAuthSignUpResult>> signup({String username, String password, String email}) async{
    final req = NUIAuthSignUpReq(username: username, password: password, email: email);

    final result = await _api.signup(request: req);
    if(result.success && result.data != null) {
      final signupToken = result.data["signupToken"] as String;
      await NUISharedPref.get().save(_PREF_SIGNUP_TOKEN, signupToken);
      return DataResult<NUIAuthSignUpResult>(success: result.success, message: result.message, data: NUIAuthSignUpResult(signupToken: signupToken));
    }
    return DataResult<NUIAuthSignUpResult>(success: result.success, message: result.message);
  }

  @override
  Future<DataResult> verifyEmail({String code}) async{
    final token = await NUISharedPref.get().getString(_PREF_SIGNUP_TOKEN);

    if (token.isNullOrEmpty()) {
      throwNUI(_MODULE, "Invalid sign up token found");
    }

    final req = NUIAuthVerifyEmailReq(code: code, token: token);
    final result = await _api.verifyEmail(request: req);

    if(result.success){
      await NUISharedPref.get().remove(_PREF_SIGNUP_TOKEN);
    }
    return DataResult(success: result.success, message: result.message);
  }
}

class NUIAuthBuilder{
  String _baseUrl;
  NUITokenExpiredListener _tokenExpiredListener;

  NUIAuthBuilder baseUrl(String baseUrl){
    _baseUrl = baseUrl;
    return this;
  }

  NUIAuthBuilder tokenExpiredListener(NUITokenExpiredListener listener){
    _tokenExpiredListener = listener;
    return this;
  }

  NUIAuth build(){
    NUIWebBuilder()
      .module(NUIAuthApi.MODULE)
      .baseURL(_baseUrl)
      .headers((){
        final accessToken = NUIAuth.get().getAccessToken();
        return {
          "Authorization": "Bearer ${accessToken?.token ?? "no-token"}",
          "Content-Type": "application/json"
        };
    })
    .preHandler((response) async{
      final accessToken = NUIAuth.get().getAccessToken();
      if (accessToken?.isExpired(leeway: 0) == true) {
        logNUI(NUIAuth._MODULE, "Access token expired, proceed to refresh token");
        final response = await NUIAuth.get()._api.refreshToken(request: NUIAuthLoginReq(grantType: NUIAuth.GRANT_TYPE_REFRESH_TOKEN, refreshToken: accessToken.refreshToken));

        if (response.success && response.data != null) {
        // save the new access token and refresh token
          final body = response.data;
          logNUI(NUIAuth._MODULE, "Token refreshed, new access token : ${body.accessToken}, refresh token : ${body.refreshToken}");
          NUIAuth.get()._saveLoginResponseData(body);
        }
        else{
          logNUI(NUIAuth._MODULE, "Failed to refresh expired token");
          if(_tokenExpiredListener != null){
            _tokenExpiredListener();
          }
        }
      }
    })
    .build();
    NUISharedPrefBuilder()
      .build();
    return NUIAuth._builder(this);
  }
}
